#include "SA_twolf.h"
#include "CellLattice.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

int nCols, nCells, nNets;
int * nets;
double T0;
long step;

CellLattice lattice;
CellLattice best;
long costBest;

#define PRINT_OK false

void SimulatedAnnealing(char * filename, double T_init,int M,double beta) {
	ReadFromFile(filename);
	if(nCols==0||nNets==0||nCells==0) {
		fprintf(stderr,"Unable to glean useful information from file: %s\n",filename);
		return;
	}

	// initialize the best to the first
	best.Init(&lattice);

/*
	lattice.dump();
	for(int i=0;i<nNets;i++) {
		printf("Net: %i\tWeight: %i\n",i+1,nets[i]);
	}
*/
	// uncomment for release
	srand(time(NULL));

	double T = T_init;
	double T_lastH = T; // used to change H
	double T_lastW = T;	// used to change W
	step = 0;
	Cell * cell;

	int H, W, H_prime, W_prime;
	int retval;

	// start horizontal and vertical search fields
	// at the complete placement area
	H = lattice.nRows();
	W = lattice.nCols() * 4;

	costBest = LatticeCost();

	// start of simulated annealing

	do{
		// Call metropolis
		retval = Metropolis(T,M,H,W);

		if(retval==-1) break;

		T = alpha(T) * T;
		M = (int) ceil(beta * M);

		// decrease search areas
		H_prime = (int) H * log(T) / log(T_lastH);
		if(H_prime<H) {
			H = H_prime;
			T_lastH = T;
			if(H<3) H = 3;
		}

		W_prime = (int) W * log(T) / log(T_lastW);
		if(W_prime<W-3) { // W must always be a power of 4
			W = W_prime & 0xFFFFFFFC;
			T_lastW = T;
			if(W<12) W = 12;
		}
	} while(T>1.0);

	fprintf(stdout,"\n");

	
	bool row_printed;
	if(retval==-1) lattice.dump();

	else {
		// do a pretty dump of best, the best solution found

		fprintf(stdout,"\n\nFinal Solution:");
		
		for(int i=0;i<best.nRows();i++) {
			if(rowPopulation(i)==0) break;

			row_printed = false;

			

			for(int j=0;j<best.nCols();j++) {
				cell = (Cell *) best[i][j];
				if(cell==NULL) continue;

				if(!row_printed) {
					retval = fprintf(stdout,"\nRow %i: ",i+1);
					for(int x=0;x<11-retval;x++) fprintf(stdout," ");
					row_printed = true;
				}

				retval = fprintf(stdout,"C%i ",cell->myNumber);
				for(int y=0;y<5-retval;y++) fprintf(stdout," ");
				
			}

		}

		fprintf(stdout,"\n\nCost of Final Solution: %i\n",costBest);
	}

	delete nets;
	lattice.FreeMem();
	best.FreeMem();
}

int Metropolis(double T, int M, int H, int W) {
	long costS, costNewS, costDelta;
	bool last_accepted = true;
	int move;
	int printed;
	int cell1, cell2;
	int r_cell1, r_cell2, c_cell1, c_cell2;
	bool disallowed;
	
	Cell * cell;
	int i, j;
	int popEff;

	do{
		// using a bool here will cut down
		// significantly on calls to LatticeCost()

		// debugging only
//		fprintf(stdout,"Before move:\n");
//		lattice.dump();

		step++;

		// pretty-print the current step
		if(PRINT_OK) {
			printed = fprintf(stdout,"%i ",step);
			for(i=0;i<8-printed;i++) fprintf(stdout," ");
		}

		if(last_accepted) costS = LatticeCost();

		// this implementation of timberwolf cannot
		// use moves of type 3--the HW definition demands
		// that orientation of modules is fixed
		move = RandOnRange(0,1);
		r_cell1 = r_cell2 = c_cell1 = c_cell2 = -1;

		switch(move) {
		case 0 :
			// Move a random single cell to a new location

			// cell2 is now a counter of the number of tries to find
			// a spot for the cell
			cell2 = -1;
			while(true) {

				cell2++;

				if(cell2>100) {
					break;
				}
				
				// select a cell to move
				cell1 = RandOnRange(1,nCells);
//				fprintf(stdout,"\nConsidering cell %i for movement:\n",cell1);

				// find the cell's location
				for( i=0;i<lattice.nRows() && (r_cell1==-1);i++) {
					for( j=0;j<lattice.nCols() && (r_cell1==-1);j++) {
						cell = (Cell *) lattice[i][j];
						if(cell==NULL) continue;
						if(cell->myNumber==cell1) {
							r_cell1 = i;
							c_cell1 = j;
						}
					}
				}

				if(r_cell1==-1) {
					fprintf(stderr,"Unable to locate cell %i\n",cell1);
					exit(-1);
					return -1;
				}

				if(cell2>0) {
					// consider doing an M3 instead of trying M1 again
					// Rong said don't do M3's so this effectively
					// disables it--rand(1,10) can never be 11
					if(RandOnRange(1,10)==11) {
						// 1 in 10 probability of doing an
						// M3 if a move is rejected (textbook)
						doM3(r_cell1,c_cell1);
						move = 2;
						break;
					}
				}

//				fprintf(stdout," Found Cell %i @ (%i,%i)\n",cell1,r_cell1,c_cell1);

			/*	if(rowPopulation(r_cell1)-1<=0) {
					r_cell1 = r_cell2 = c_cell1 = c_cell2 = -1;
					//fprintf(stdout,"D-RPC, effPop = %i\n",rowPopulation(r_cell1)-1);
					continue;
				} */

				// find a random unoccupied destination inside the movement range

//				fprintf(stdout,"Looking for a spot for Cell %i\n",cell1);

				r_cell2 = r_cell1+RandOnRange(-W/4,W/4);
				while(r_cell2>=lattice.nRows() || r_cell2<0) 
					r_cell2 = r_cell1+RandOnRange(-W/4,W/4);

				if(rowPopulation(r_cell2)==lattice.nCols()) {// no room in the row
					r_cell1 = r_cell2 = c_cell1 = c_cell2 = -1;
					continue;	
				}

				c_cell2 = c_cell1+RandOnRange(-H,H);
				costNewS = 0;
				while(c_cell2>=lattice.nCols() || c_cell2<0||c_cell2>rowPopulation(r_cell2)) {
					c_cell2 = c_cell1+RandOnRange(-H,H);
					costNewS++;
					if(costNewS>100) break;
				}

				if(costNewS>100) continue;

				// there must never be an empty row separating groups
				// of cells
				disallowed = false; 

				// now check to make sure the first row is never empty
				if(r_cell1==0 && rowPopulation(0)==1) {
					disallowed = true;
				}

				// now check to make sure no rows are skipped
				if(r_cell2>0 && rowPopulation(r_cell2-1)==0) {
					disallowed = true;
				}


				// check the row that is being moved out of to see if
				// it separates two groups of rows
				popEff = rowPopulation(r_cell1) - 1;
				if(popEff<=0) {
					// its still OK if all rows below this row
					// are empty
					for(i=r_cell1+1;i<lattice.nRows();i++) {
						if(rowPopulation(i)+((i==r_cell2) ? 1:0)>0) {
							disallowed = true;
						}
					}
				}

				// regardless of whether a spot is occupied, make room if 
				// possible
				if(!disallowed && makeRoom(r_cell2,c_cell2)) {
					// room has been made, move the cell
					lattice[r_cell2][c_cell2] = lattice[r_cell1][c_cell1];
					lattice[r_cell1][c_cell1] = NULL;

//					fprintf(stdout,"Moving cell %i from (%i,%i)->(%i,%i)\n",
//						cell1,r_cell1,c_cell1,r_cell2,c_cell2);

					// fill any holes generated by moving out of row
					// r_cell1
					fillHole(r_cell1);
				} else {
					// couldnt' make room in that row
					disallowed= true;
				}

								
				if(disallowed) {
					r_cell1 = r_cell2 = c_cell1 = c_cell2 = -1;
					continue;
				} 

//				fprintf(stdout,"Cell %i now at (%i,%i)\n",cell1,r_cell2,c_cell2);

				break;			

			}

			if(cell2<=100) {
				break;
			}

			// allow fall-through if (cell2=tries)>100

		case 1 :
		default:
			move = 1;
			// swap two cells

			// first find two cells inside of swapping range
			while(true) {
				r_cell2 = r_cell1 = c_cell1 = c_cell2 = -1;

				// Try random cells
				cell1 = RandOnRange(1,nCells);
				cell2 = RandOnRange(1,nCells);
//				fprintf(stdout,"\nLooking for cells %i and %i:\n",cell1,cell2);

				// search the matrix for the cells
				for( i=0;i<lattice.nRows() && (r_cell2==-1||r_cell1==-1);i++) {
					for( j=0;j<lattice.nCols() && (r_cell2==-1||r_cell1==-1);j++) {
						
						cell = (Cell *) lattice[i][j];
						if(cell==NULL) continue;

						if(cell->myNumber==cell1) {
//							fprintf(stdout,"Found cell %i @ (%i,%i)\n",cell1,i,j);
							r_cell1 = i;
							c_cell1 = j;
						}

						if(cell->myNumber==cell2) {
//							fprintf(stdout,"Found cell %i @ (%i,%i)\n",cell2,i,j);
							r_cell2 = i;
							c_cell2 = j;
						}
					}
				}
	
				if(r_cell1==-1||r_cell2==-1) {
					fprintf(stderr,"Unable to locate cells %i and %i\n",cell1, cell2);
					exit(-1);
					return -1;
				}	

				// check distance parameters
				if(abs(r_cell1-r_cell2)<=H && abs(c_cell1-c_cell2)<=W/4)
					break;
			}
		

			// Swap the two cells

			// using costNewS as a temporary long here

//			fprintf(stdout,"Swaping Cell %i (%i,%i) with Cell %i (%i,%i)\n",
//				cell1,r_cell1,c_cell1,
//				cell2,r_cell2,c_cell2);
			costNewS = lattice[r_cell1][c_cell1];
			lattice[r_cell1][c_cell1] = lattice[r_cell2][c_cell2];
			lattice[r_cell2][c_cell2] = costNewS;

			// swap complete

			break;
		}
		// pretty-print the perturbation
		if(PRINT_OK) {
		switch(move) {
		case 0 :
			fprintf(stdout,"M1 ");
			break;
		case 1:
			fprintf(stdout,"M2 ");
			break;
		case 2:
		default:
			fprintf(stdout,"M3 ");
		}
		}

		costNewS = LatticeCost();

		if(PRINT_OK){
		// pretty-print the cost of the perturbation
		printed = fprintf(stdout,"%i ",costNewS);
		for(i=0;i<10-printed;i++) fprintf(stdout," ");

		// pretty-print the temperature
		printed = fprintf(stdout,"%f ",T);
		for(i=0;i<17-printed;i++) fprintf(stdout," ");
		}

		if(costNewS<costBest) {
			// best solution seen so far
			best.FreeMem();
			best.Init(&lattice);
			costBest = costNewS;
			if(best.nCols()==0) {
				fprintf(stderr,"Step %i\tM%i\tLatticeR=%i\n",step,move+1,lattice.nRows());
				lattice.dump();
				exit(-1);
				// todo todo todo TODO
				// Test this
			}
		}

		costDelta = costNewS - costS;

//		fprintf(stdout,"\nAfter move:\n");
//		lattice.dump();

		if(costDelta<0) {
			// accept the solution, so don't undo the change
			last_accepted = true;

			// print acceptance
			if(PRINT_OK) fprintf(stdout,"Y\n");
		} else if(Rand01()<exp(-costDelta/T)) {
			// accept an uphill move, don't undo the change
			last_accepted = true;
			
			// print acceptance
			if(PRINT_OK) fprintf(stdout,"Y\n");
		} else {
			// do not accept the move
			last_accepted = false;

			// print rejection
			if(PRINT_OK) fprintf(stdout,"N\n");

			// undo the move here
			switch(move) {
			case 0 :
				// undo a single-cell move here

				// this cannot return false, since the cell
				// that left this spot by default left room
				makeRoom(r_cell1,c_cell1);
				lattice[r_cell1][c_cell1] = lattice[r_cell2][c_cell2];
				lattice[r_cell2][c_cell2] = NULL;
				fillHole(r_cell2);

				break;
			case 1 :
				// undo a swap here
				// using costNewS as a temporary long here
//				lattice.dump();

//				fprintf(stdout,"Attempting to undo an M2:\n");
//				fprintf(stdout,"Swap was (%i,%i)<->(%i,%i)\n",
//					r_cell1,c_cell1, r_cell2,c_cell2);
				costNewS = lattice[r_cell1][c_cell1];
				lattice[r_cell1][c_cell1] = lattice[r_cell2][c_cell2];
				lattice[r_cell2][c_cell2] = costNewS;
				
//				lattice.dump();
				break;
			case 2:
			default:
				// undo a mirroring here
				doM3(r_cell1,c_cell1);
				break;
			}
		}

		M = M - 1;
	} while(M>0);

	return 0;
}

void doM3(int row, int col) {
	Cell * cell;

	cell = (Cell *) lattice[row][col];
	if(cell==NULL) return; // do nothing, but never happens

	cell->MirrorAboutX();
}

// 100,10,10,100 for 10 cells
// 10 10 10 2500 for 50 cells & 10 cells
// 10 10 10 2500

// for comparison to genetic, use 0.0, 0.0, 1.0, 0.0
#define IMBALANCE_WEIGHT 0.0
#define BOUNDING_AREA_WEIGHT 0.0
#define VERTICAL_INTERCONNECT_WEIGHT 1.0
#define EFF_ROWS_WEIGHT 0.0

long LatticeCost() {
	int i;

	// I will introduce another element of cost into this cost function:
	// the area of the minimum bounding rectangle
	long bounding_area;

	// estimate the interconnect cost
	long interconnect = 0;

	for(i=0;i<nNets;i++) {
		interconnect += nets[i] * findSpan(i+1);
	}

	// Now,
	// measure the row imbalance
	long imbalance;

	long rowlen_average;
	long nRowsEff = 0;
	long rowlen;
	long rowlen_max = 0;

	// first compute rowlen_average
	rowlen_average = 0;
	for(i=0;i<lattice.nRows();i++) {
		rowlen = rowPopulation(i);
		if(rowlen_max<rowlen) rowlen_max = rowlen;
		if(rowlen!=0) {
			rowlen_average += rowlen;
			nRowsEff++;
		}
	}
	if(nRowsEff!=0) {
		rowlen_average = rowlen_average / nRowsEff;
	}

	// now compute the weigthed sum of the deviations
	imbalance = 0;

	for(i=0;i<lattice.nRows();i++) {
		imbalance += abs(rowPopulation(i) - rowlen_average);
	}

	imbalance = (long) IMBALANCE_WEIGHT * imbalance;

//	fprintf(stdout,"***Imbalance = %i\tInterconnect = %i***\n",imbalance,interconnect);
	
	// include the bounding area
	bounding_area = (long) BOUNDING_AREA_WEIGHT * 
					(4*nRowsEff) * (4*rowlen_max);

	// include the number of rows in the cost function
	nRowsEff = (long) (nCells>20 ? 5*EFF_ROWS_WEIGHT : EFF_ROWS_WEIGHT) * nRowsEff;

	return imbalance + interconnect + bounding_area + nRowsEff;
}

int findSpan(int net) {
	int uppermost = 4*(lattice.nRows()+10);
	int lowermost = -1;
	int leftmost = 4*(lattice.nCols()+10);
	int rightmost = -1;

	int x, y;

	NetConnection connection;

	Cell * cell;

	for(int i=0;i<lattice.nRows();i++) {
		for(int j=0;j<lattice.nCols();j++) {
			cell = (Cell *) lattice[i][j];
			if(cell==NULL) continue;

			connection = cell->getConnection(net);
			if(connection==not_connected) continue;

			// it will be connected to something
			// for comparision to genetic, use average cell position
			x = 4*j+2;
			y = 4*i+2;

			/*
			if(connection==ul_port) {
				y = 4*i;
				x = 4*j+1;

			} else
			if(connection==um_port) {
				y = 4*i;
				x = 4*j+2;

			} else
			if(connection==ur_port) {
				y = 4*i;
				x = 4*j+3;

			} else
			if(connection==ll_port) {
				y = 4*i+4;
				x = 4*j+1;

			} else
			if(connection==lm_port) {
				y = 4*i+4;
				x = 4*j+2;

			} else {
				// must be connected to lr_port
				y = 4*i+4;
				x = 4*j+3;
			}
			// */
			if(y>lowermost) lowermost = y;
			if(y<uppermost) uppermost = y;
			if(x>rightmost) rightmost = x;
			if(x<leftmost)  leftmost  = x;
			
		}
	}

	
//	fprintf(stdout,"Vertical Span = %i\tHorizont Span = %i\n",lowermost-uppermost,rightmost-leftmost);


	return (rightmost-leftmost) + ((long) VERTICAL_INTERCONNECT_WEIGHT*(lowermost-uppermost));
}

int rowPopulation(int row) {
	int pop = 0;
	if(row<0||row>=lattice.nRows()) return -1;
	for(int j=0;j<lattice.nCols();j++) {
		if(lattice[row][j]!=NULL) pop++;
	}
	return pop;
}

bool makeRoom(int row, int col) {
	if(row<0||row>=lattice.nRows()||col<0||col>=lattice.nCols()) return false;
	
	// first check to see if room must be made, or if the spot
	// is already empty
	if(lattice[row][col]==NULL) return true;

	// now check to see if room can be made, that is, the row is
	// not full
	int pop = rowPopulation(row);
	if(pop==lattice.nCols()) return false;

	// so room must exist at the end of the row
	// so move everything over

//	fprintf(stdout,"Before making room @ (%i,%i):\n",row,col);
//	lattice.dump();

	int j;
	for(j=lattice.nCols()-2;j>=col;j--) {
		lattice[row][j+1] = lattice[row][j];
	}

	lattice[row][col] = NULL;

//	fprintf(stdout,"After making room:\n");
//	lattice.dump();

	return true;
	
}

// this is faithfully called after every single-cell move--
// therefore AT MOST ONE hole must be filled
void fillHole(int row) {
	int pop = rowPopulation(row);
	if(pop==0||pop==-1) return; // no holes to fill

//	fprintf(stdout,"\nThis is the lattice as fillHole(%i) sees it:\n",row);
//	lattice.dump();

	// find the hole, if it exists
	int j; int hole =-1;
	for(j=0;j<lattice.nCols()-1;j++) {
		if(lattice[row][j]==NULL) {
			if(lattice[row][j+1]!=NULL) {
				hole = j;
				break;
			} else {
				return; //  no hole exits, just end of the row
			}
		}
	}
	if(hole==-1) return; // there is no hole

//	fprintf(stdout,"Before filling hole @ col %i\n",hole);
//	lattice.dump();

	// scoot everything on this row down once
	for(j=hole+1;j<lattice.nCols();j++) {
		lattice[row][j-1] = lattice[row][j];
	}

	lattice[row][lattice.nCols()-1] = NULL; // null-terminate

//	fprintf(stdout,"After filling hole:\n");
//	lattice.dump();
}

double alpha(double T) {
	if(T> 0.67*T0) return 0.8;
	if(T> 0.33*T0) return 0.95;
	
	return 0.85;
}

void ReadFromFile(char * filename) {
	GetMetaData(filename);
	if(nCols==0||nNets==0||nCells==0) {
		return;
	}

	char buff_in[1024];
	char * token;

	int current_net = 0;
	int number, one, two, three, four, five;
	Cell * temp;

	// allocate space for the lattice and nets
	nets = new int[nNets];
	lattice.Init(nCols,nCells);

	FILE * infile = fopen(filename,"r");

	// skip the first line
	fgets(buff_in,1024,infile);

	// process all remaining lines
	while(fgets(buff_in,1024,infile)) {
		token = strtok(buff_in," \t\n");
		if(token==NULL) continue;
		if(token[0]=='N') {
			// found a net
			if(atoi(&token[1])-1!=current_net) {
				fprintf(stderr,"Danger Will Robinson! Nets not defined in order\n");
			}
			token = strtok(NULL," \t\n");
			nets[current_net] = atoi(token);
			current_net++;
		} else if(token[0]=='C') {
			// found a cell
			number = atoi(&token[1]);
			token = strtok(NULL," \t\n");
			one = atoi(&token[1]);
			token = strtok(NULL," \t\n");
			two = atoi(&token[1]);

			// may have up to three more
			token = strtok(NULL," \t\n");
			if(token!=NULL) {
				// has 3 or more net connections
				three = atoi(&token[1]);

				token = strtok(NULL," \t\n");
				if(token!=NULL) {
					// has 4 or more net connections
					four = atoi(&token[1]);

					token = strtok(NULL," \t\n");
					if(token!=NULL) {
						// has 5 net connections
						five = atoi(&token[1]);

						temp = new Cell(number,one,two,three,four,five);
					} else {
						// has 4 net connections
						temp = new Cell(number,one,two,three,four);
					}
				} else {
					// has 3 net connections
					temp = new Cell(number,one,two, three);
				}
			} else {
				// has 2 net connections
				temp = new Cell(number,one,two);
			}

			lattice.AddCell(temp);

		} else continue;
	}

	fclose(infile);
}

void GetMetaData(char * filename) {
	nCols = 0;
	nCells = 0;
	nNets = 0;
	
	int net, cell;

	FILE * infile = fopen(filename,"r");
	if(infile==NULL) return;
	
	char buff_in[1024];
	fgets(buff_in,1024,infile);
	char * token;

	// get max row length (nCols)
	token = strtok(buff_in," \t\n");
	nCols = atoi(token);
	nCols = nCols / 4;

	while(fgets(buff_in,1024,infile)) {
		token = strtok(buff_in," \n\t");
		if(token==NULL) continue; // ignore blank lines
		if(token[0]=='N') {
			// found a net
			net = atoi(&token[1]);
			if(net>nNets) nNets = net;

			continue;
			
		} else if(token[0]=='C') {
			// found a cell
			cell = atoi(&token[1]);
			if(cell>nCells) nCells = cell;
		} else {
			continue;
		}
	}

	fclose(infile);


}

int RandOnRange(int low, int high) {
	if(low>high) return 0;
	return (rand()%(high-low+1))+low;
	
};

double Rand01() {
	double retval;

	retval = 0.001 * RandOnRange(0,999);

	return retval;

}

